from .keapsession import KeapSession


class KeapRestClient(KeapSession):

    def __init__(self, app):
        KeapSession.__init__(self, app)

    def validate_auth(self):
        return self.is_authorized()

    def make_request(self, method, endpoint, data=None, json_data=None, **kwargs):
        if not self.is_authorized():
            raise KeapAuthError("Keap is not authorized.")

        url = f"{self.api_base_url}{endpoint}"
        attempts = 0
        while attempts < 3:
            if method == "get":
                response = self.request(method, url, params=kwargs)
            else:
                response = self.request(method, url, data=data, json_data=json_data)
            if response.status_code == 200:
                return response.json()
            elif response.status_code == 204 or 201:
                return True
            elif response.status_code == 401:
                self.refresh_session_auth()
                attempts += 1
            else:
                self.handle_keap_error(response)
                attempts = 3

    def _get(self, endpoint, data=None, **kwargs):
        return self.make_request('get', endpoint, data=data, **kwargs)

    def _post(self, endpoint, data=None, json_data=None, **kwargs):
        return self.make_request('post', endpoint, data=data, json_data=json_data, **kwargs)

    def _delete(self, endpoint, **kwargs):
        return self.make_request('delete', endpoint, **kwargs)

    def _patch(self, endpoint, data=None, json_data=None, **kwargs):
        return self.make_request('patch', endpoint, data=data, json_data=json_data, **kwargs)

    def _put(self, endpoint, json_data=None, **kwargs):
        return self.make_request('put', endpoint, json_data=json_data, **kwargs)

    @staticmethod
    def handle_keap_error(response):
        """
            This method get the response request and returns json_data data or raise exceptions
            :param response:
            :return:
        """
        if response.status_code == 400:
            raise KeapDataError(
                "The URL {0} retrieved an {1} error. "
                "Please check your request body and try again.\nRaw message: {2}".format(
                    response.url,
                    response.status_code,
                    response.text
                )
            )
        elif response.status_code == 401:

            raise KeapAuthError(
                "The URL {0} retrieved and {1} error. Please check your credentials, "
                "make sure you have permission to perform this action and try again.".format(
                    response.url,
                    response.status_code
                )
            )
        elif response.status_code == 403:
            raise KeapAuthError(
                "The URL {0} retrieved and {1} error. This action was forbidden.".format(
                    response.url,
                    response.status_code
                )
            )
        elif response.status_code == 404:
            raise ConnectionError(
                "The URL {0} retrieved an {1} error. Please check the URL and try again.\n"
                "Raw message: {2}".format(
                    response.url,
                    response.status_code,
                    response.text
                )
            )
        else:
            raise KeapException(
                "An Unknown Issue occurred"
            )


class KeapException(Exception):
    """Base exception for all library errors."""
    pass


class KeapAuthError(KeapException):
    """Something is wrong with authentication."""
    pass


class KeapTokenError(KeapAuthError):
    """Something is wrong with the tokens."""
    pass


class KeapDataError(KeapException, ValueError):
    """Something is wrong with the data."""
    pass


class KeapConnectionError(KeapException, ConnectionError):
    """Something is wrong with the connection."""
    pass



import pickle


class KeapCache:

    def __init__(self, cache_file=None):
        self.cache_file = cache_file
        self.cache = self.get_dict('keap_cache')

    def refresh_cache(self):
        self.cache = self.get_dict('keap_cache')

    def update_cache(self, key, data):
        self.refresh_cache()
        self.cache[key] = data
        self.update_dict_data("keap_cache", self.cache)

    def update_dict_data(self, dict_name, data):
        cache = self.read_all_data().copy()
        if dict_name in cache.keys():
            cache[dict_name].update(data)
        else:
            cache[dict_name] = data
        # if dict_name in current_cache:
        pickle_out = open(self.cache_file, "wb")
        pickle.dump(cache, pickle_out)
        pickle_out.close()

    def read_all_data(self):
        try:
            pickle_in = open(self.cache_file, "rb")
            data = pickle.load(pickle_in)
            pickle_in.close()
        except FileNotFoundError:
            data = {}
            pickle_out = open(self.cache_file, "wb")
            pickle.dump(data, pickle_out)
            pickle_out.close()
        return data

    def get_dict(self, dict_name):
        data = self.read_all_data().get(dict_name)
        if data is None:
            self.update_dict_data(dict_name, {})
            return self.read_all_data().get(dict_name)
        else:
            return data



import logging
import os
import time
from datetime import datetime
from json import dump, load
from logging.handlers import RotatingFileHandler
from os import environ
from os.path import join
from xmlrpc.client import ServerProxy, Error
from dateutil import parser
from .keapcache import KeapCache
from authlib.integrations.requests_client import OAuth2Session
from flask import Blueprint, request, redirect, url_for


class KeapSession(OAuth2Session, KeapCache):

    def __init__(self, app, auth_class=None, logger=None, cache=None):
        APP_BASE_DIR = app.root_path
        if auth_class is None:
            # self.auth_handler = JsonAuthHandler((join(str(Path(__file__).parent.parent.parent.resolve()), "keys/keap_token.json")))
            if 'KEYS_DIR' not in app.config.keys():
                keys_dir = app.root_path
            else:
                keys_dir = app.config.get('KEYS_DIR')
            self.auth_handler = JsonAuthHandler(join(keys_dir, 'keap_token.json'))
        else:
            self.auth_handler = auth_class()

        if logger is None:
            self.logger = self.setup_logger(APP_BASE_DIR)
        else:
            self.logger = logger

        self.token_endpoint = "https://api.infusionsoft.com/token"
        self.auth_endpoint = 'https://signin.infusionsoft.com/app/oauth/authorize'
        self.api_base_url = 'https://api.infusionsoft.com/crm/rest/v1/'
        OAuth2Session.__init__(self, token=self.auth_handler.get_token(), client_id=app.config.get('KEAP_CLIENT_ID'),
                               client_secret=app.config.get('KEAP_CLIENT_SECRET'),
                               update_token=self.auth_handler.store_token,
                               redirect_uri=f"{app.config.get('AUTH_APP_BASE_URL')}authorize-keap",
                               token_endpoint=self.token_endpoint)
        cache_dir = (os.path.join(APP_BASE_DIR, "cache"))
        cache_file = (os.path.join(cache_dir, "cache"))
        KeapCache.__init__(self, cache_file=cache_file)
        self.contact_custom_fields = None
        self.tags = None
        self.name = "Keap-"
        app.register_blueprint(self.get_blueprint())

    def refresh_session_auth(self):
        try:
            self.refresh_token(url=self.token_endpoint)
            self.logger.info("Token Refreshed.")
            return True
        except Exception as e:
            self.logger.warning("Token Refresh Failed.")
            self.logger.exception(e)
            return False

    def update_authorization(self, authorization_data, **kwargs):
        self.auth_handler.update_authorization(authorization_data)

    def is_auth_expired(self):
        return self.token.is_expired()

    def has_token(self):
        if self.token is None:
            return False
        else:
            return True

    def is_authorized(self):
        if not self.has_token():
            return False
        if self.is_auth_expired():
            return self.refresh_session_auth()
        else:
            return True

    def validate_auth(self):
        return self.is_authorized()

    def get_xmlrpc_client(self):
        if not self.is_authorized():
            return None
        else:
            return InfusionsoftOAuth(self.token.get("access_token"))

    def get_authorization_url(self):
        return self.create_authorization_url(self.auth_endpoint)

    def handle_callback(self, callback_request, code):
        token = self.fetch_token(url=self.token_endpoint, authorization_response=callback_request.url, scope='full',
                                 client_id=self.client_id,
                                 redirect_uri=self.redirect_uri,
                                 client_secret=self.client_secret, code=code,
                                 grant_type='authorization_code')
        self.auth_handler.store_token(token)

    def get_blueprint(self):
        keap_blueprint = Blueprint('keap_blueprint', __name__, cli_group="keap")

        @keap_blueprint.route('/authorize-keap')
        def authorize_keap():
            code = request.args.get('code', None)
            if not code:
                uri, state = self.get_authorization_url()
                return redirect(uri, code=302)
            else:
                self.handle_callback(request, code=code)
                self.set_tags()
                self.get_contact_custom_fields()
                return redirect(url_for('index'))

        @keap_blueprint.cli.command()
        def refresh_auth():
            self.refresh_session_auth()

        return keap_blueprint

    @staticmethod
    def get_field_map():
        return CustomFieldMap.keap

    def set_tags(self):
        tags = {}
        tag_names = environ.get('KEAP_TAGS').split(',')
        for tag_name in tag_names:
            tag_id = self.get_tag_id(tag_name)
            if tag_id is not None:
                tags[tag_name] = tag_id
            else:
                continue
        self.update_cache('tags', tags)

    def get_tag_id(self, tag_name):
        params = {"name": tag_name}
        response = self.session.get(f'{self.api_base_url}tags', params=params)
        if response.json().get('count') > 0:
            return response.json().get('tags')[0].get('id')
        else:
            return None

    def get_contact_custom_fields(self):
        if self.contact_custom_fields is None:
            response = self.session.get(f'{self.api_base_url}contactCustomFields')
            if response.status_code == 200:
                custom_fields = response.json()
                keap_custom_fields = KeapCustomFields()
                for field in custom_fields:
                    keap_custom_fields[f"_{field.get('field_name')}"] = field.get("id")
                self.contact_custom_fields = keap_custom_fields
                return keap_custom_fields
            else:
                return None
        else:
            return self.contact_custom_fields

    def get_contact_custom_field_id(self, field_key):
        fields = self.get_contact_custom_fields()
        keap_field_name = CustomFieldMap.keap.get(f"{field_key}")
        return fields.get(f"{keap_field_name}")

    @staticmethod
    def setup_logger(base_dir):
        LOG_DIR = (os.path.join(base_dir, "logs"))
        LOG_FORMATTER = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        log_file_handler = RotatingFileHandler(LOG_DIR + f'/keap_client.log', maxBytes=100000, backupCount=5)
        log_file_handler.setFormatter(LOG_FORMATTER)
        if not os.path.exists(LOG_DIR):
            try:
                os.makedirs(LOG_DIR)
            except Exception as e:
                print(e)
        logger = logging.getLogger("keap_client")
        logger.addHandler(log_file_handler)
        logger.setLevel(logging.DEBUG)
        return logger




class Infusionsoft(object):
    base_uri = 'https://%s.infusionsoft.com/api/xmlrpc'

    def __init__(self, name, api_key, use_datetime=False):
        uri = self.base_uri % name
        self.client = ServerProxy(uri, use_datetime=use_datetime)
        self.client.error = Error
        self.key = api_key

    def __getattr__(self, service):
        def function(method, *args):
            call = getattr(self.client, service + '.' + method)

            try:
                return call(self.key, *args)
            except self.client.error as v:
                return "ERROR", v

        return function

    def server(self):
        return self.client


class InfusionsoftOAuth(Infusionsoft):
    base_uri = 'https://api.infusionsoft.com/crm/xmlrpc/v1?'

    def __init__(self, access_token, use_datetime=False):
        uri = '%saccess_token=%s' % (self.base_uri, access_token)

        self.client = ServerProxy(uri, use_datetime=use_datetime, allow_none=True)
        self.client.error = Error
        self.key = access_token


class JsonAuthHandler:

    def __init__(self, auth_file_path):
        self.auth_file_path = auth_file_path

    def store_token(self, data, **kwargs):
        data["updated_at"] = time.time()
        if "expires_at" in data:
            date = datetime.fromtimestamp(data.get("expires_at"))
            data["expiration_datetime"] = date.strftime("%m-%d-%Y  %H:%M")
        else:
            data["expiration_datetime"] = 0
        with open(self.auth_file_path, "w", encoding="utf-8") as f:
            dump(data, f, ensure_ascii=False, indent=4)

    def get_token(self):
        try:
            with open(self.auth_file_path, "r", encoding="utf-8") as f:
                data = load(f)
                if data == {}:
                    return None
                return data
        except FileNotFoundError:
            data = {}
            with open(self.auth_file_path, "w", encoding="utf-8") as f:
                dump(data, f, ensure_ascii=False, indent=4)
            return None

    def get_authorization(self):
        return self.get_token()

    def update_authorization(self, authorization_data, **kwargs):
        updated_auth = self.get_token()
        updated_auth["access_token"] = authorization_data["access_token"]
        updated_auth["refresh_token"] = authorization_data["refresh_token"]
        updated_auth["expires_at"] = datetime.timestamp(parser.parse(authorization_data["expires_at"]))
        self.store_token(updated_auth)




class KeapCustomFields(dict):
    pass


class CustomFieldMap:
    keap = {}
    keap_goals = {}
    for key, value in environ.items():
        if "IFS" in key and "FIELD" in key:
            keap[key] = value
        if "KEAP" in key and "FIELD" in key:
            keap[key] = value
        if "KEAP" in key and "GOAL" in key:
            keap_goals[key] = value
    pass



from json import dumps

from .keapcache import KeapCache
from .keaprestclient import KeapRestClient, KeapDataError


class KeapClient(KeapRestClient, KeapCache):

    def __init__(self, app):
        KeapRestClient.__init__(self, app)

    def update_contact_custom_field(self, contact_id, field_key, value):
        field_id = self.get_contact_custom_field_id(field_key=field_key)
        custom_fields = [{
            "id": f"{field_id}",
            "content": f"{value}"
        }]
        data = {
            "custom_fields": custom_fields
        }
        self.update_contact(contact_id, **data)

    def get_contacts(self, **kwargs):
        """
            To get all the contacts you can just call the method, to filter use limit, order, offset.
            For other options see the documentation of the API
            :return:
        """
        return self._get('contacts', **kwargs)

    def get_contact_by_email(self, email):
        response = self.get_contacts(**{"Email": email})
        if response.get('count') == 0:
            return None
        else:
            return response.get("contacts")

    def contact_exists(self, email):
        """
            returns true if contact with given email exists
            :param email:
            :return:
        """
        count = self.get_contacts(**{"Email": email}).get("count")
        if count > 0:
            return True
        else:
            return False

    def retrieve_contact(self, id, **kwargs):
        if not id:
            raise KeapDataError("The ID is required.")

        endpoint = 'contacts/{0}'.format(id)
        return self._get(endpoint, **kwargs)

    def create_contact(self, **kwargs):
        """
            For create a contact is obligatory to fill the email or the phone number.
            I also recommend to fill the given_name="YOUR NAME"
            :param email:
            :param phone_number:
            :param kwargs:
            :return:
        """
        if kwargs is not None:
            params = {}
            params.update(kwargs)
            return self._post('contacts', json_data=params)
        raise KeapDataError("To create a contact a valid name and email is necessary")

    def add_note(self, contact_id, note_title, note_body):
        params = {
            "contact_id": contact_id,
            "title": note_title,
            "body": note_body,
            "type": "Other"
        }

        return self._post('notes', json_data=params)

    def addwithdupecheck(self, **kwargs):
        """
            For create a contact is obligatory to fill the email or the phone number.
            I also recommend to fill the given_name="YOUR NAME"
            :param email:
            :param kwargs:
            :return:
        """
        if kwargs is not None:
            params = {'duplicate_option': 'EmailAndName'}
            params.update(kwargs)
            return self._put('contacts', json_data=params)
        raise KeapDataError("Email required to check for duplicates")

    def delete_contact(self, id):
        """
            To delete a contact is obligatory send The ID of the contact to delete
            :param id:
            :return:
        """
        if not id:
            raise KeapDataError("The ID is required.")

        endpoint = 'contacts/{0}'.format(id)
        return self._delete(endpoint)

    def update_contact(self, id, **kwargs):
        """
            To update a contact you must to send The ID of the contact to update
            For other options see the documentation of the API
            :param id:
            :param kwargs:
            :return:
        """
        if not id:
            raise KeapDataError("The ID is required.")

        params = {}
        endpoint = 'contacts/{0}'.format(id)
        params.update(kwargs)
        return self._patch(endpoint, json_data=params)

    def get_campaigns(self, **kwargs):
        """
            To get the campaigns just call the method or send options to filter
            For more options see the documentation of the API
            :return:
        """
        return self._get('campaigns', **kwargs)

    def retrieve_campaign(self, id, **kwargs):
        """
            To retrieve a campaign is necessary the campaign id
            For more options see the documentation of the API
            :param id:
            :param kwargs:
            :return:
        """
        if not id:
            raise KeapDataError("The ID is required.")

        endpoint = 'campaigns/{0}'.format(id)
        return self._get(endpoint, **kwargs)

    def get_emails(self, **kwargs):
        """
            To get the emails just call the method, if you need filter options see the documentation of the API.
            :param limit:
            :param offset:
            :param kwargs:
            :return:
        """
        return self._get('emails', **kwargs)

    def get_opportunities(self, **kwargs):
        """
            To get the opportunities you can just call the method.
            Also you can filter, see the options in the documentation API.
            :param limit:
            :param order:
            :param offset:
            :param kwargs:
            :return:
        """
        return self._get('opportunities', **kwargs)

    def get_opportunities_pipeline(self):
        """
            This method will return a pipeline of opportunities
            :return:
        """
        return self._get('opportunity/stage_pipeline')

    def get_opportunity_custom_fields(self):
        return self._get('opportunities/model')

    def get_opportunity_custom_field_id(self, field_name):
        custom_fields = self.get_opportunity_custom_fields().get('custom_fields')
        field_id = None
        for field in custom_fields:
            if field['field_name'] == field_name:
                field_id = field['id']

        return field_id

    def retrieve_opportunity(self, opp_id, **kwargs):
        """
            To retrieve a campaign is necessary the campaign id
            For more options see the documentation of the API
            :param opp_id:
            :return:
        """
        if not id:
            raise KeapDataError("The ID is required.")

        endpoint = 'opportunities/{0}'.format(opp_id)
        return self._get(endpoint, kwargs)

    def create_opportunity(self, **kwargs):
        """
            To create an opportunity is obligatory to send a title of the opportunity,
            the contact who have the opportunity, and stage.
            For more information see the documentation of the API.
            :param kwargs:
            :return:
        """
        if kwargs:
            params = {}
            params.update(kwargs)
            return self._post('opportunities', json_data=params)

    def update_opportunity(self, id, **kwargs):
        """
            To update an opportunity is obligatory The ID, the other fields you can see in the documentation.
            :param id:
            :param kwargs:
            :return:
        """
        if not id:
            raise KeapDataError("The ID is required.")

        params = {}
        endpoint = 'opportunities/{0}'.format(id)
        params.update(kwargs)
        return self._patch(endpoint, json_data=params)

    def get_products(self, **kwargs):
        return self._get('products/search', **kwargs)

    def retrieve_product(self, id):
        if not id:
            raise KeapDataError("The ID is required.")

        endpoint = "products/{0}".format(id)
        return self._get(endpoint)

    def get_tasks(self, **kwargs):
        return self._get('tasks', **kwargs)

    def create_task(self, **kwargs):
        if not kwargs:
            raise KeapDataError("title and a due_date are required to create a task.")

        params = {}
        params.update(kwargs)
        return self._post('tasks', json_data=params)

    def delete_task(self, id):
        if not id:
            raise KeapDataError("The ID is required.")

        endpoint = 'tasks/{0}'.format(id)
        return self._delete(endpoint)

    def update_task(self, id, **kwargs):
        if not id:
            raise KeapDataError("The ID is required.")

        params = {}
        endpoint = 'tasks/{0}'.format(id)
        params.update(kwargs)
        return self._patch(endpoint, json_data=params)

    def retrieve_task(self, id):
        if not id:
            raise KeapDataError("The ID is required.")

        endpoint = "tasks/{0}".format(id)
        return self._get(endpoint)

    def replace_task(self, id, **kwargs):
        if not id:
            raise KeapDataError("The ID is required.")

        endpoint = "tasks/{0}".format(id)
        return self._put(endpoint, **kwargs)

    def get_orders(self, **kwargs):
        return self._get('orders', **kwargs)

    def retrieve_order(self, id):
        if not id:
            raise KeapDataError("The ID is required.")

        endpoint = "tasks/{0}".format(id)
        return self._get(endpoint)

    def get_hook_events(self):
        callback = "{0}/{1}".format("hooks", "event_keys")
        return self._get(callback)

    def get_hook_subscriptions(self):
        return self._get('hooks')

    def verify_hook_subscription(self, id):
        if not id:
            raise KeapDataError("The ID is required.")
        callback = "{0}/{1}/{2}".format("hooks", id, "verify")
        return self._post(callback, data=None)

    def create_hook_subscription(self, event, callback):
        if not event:
            raise KeapDataError("The event is required.")
        if not callback:
            raise KeapDataError("The callback is required.")

        args = {"eventKey": event, "hookUrl": callback}
        return self._post('hooks', json_data=args)

    def update_hook_subscription(self, id, event, url):
        if not id:
            raise KeapDataError("The ID is required.")

        callback = "{0}/{1}".format("hooks", id)
        args = {"eventKey": event, "hookUrl": url}
        return self._post(callback, json_data=args)

    def delete_hook_subscription(self, id):
        if not id:
            raise KeapDataError("The ID is required.")

        callback = "{0}/{1}".format("hooks", id)
        return self._delete(callback)

    def apply_tag(self, tag_name: str, contact_ids, **kwargs):
        """
        Apply a tag to one or multiple contacts.
        :param tag_name: String, tag name.
        :param contact_ids: Either one contact ID or a list of contact IDs.
        :param kwargs:
        :return:
        """
        if not tag_name:
            raise KeapDataError("tag_name is required.")

        if not isinstance(contact_ids, (tuple, list, set)) or not contact_ids:
            if contact_ids:
                contact_ids = [contact_ids]
            else:
                raise KeapDataError("contact_ids has to be a list or a single contact id.")
        tag_id = self.tags.get(tag_name)

        data = {
            "ids": contact_ids
        }
        endpoint = 'tags/{tagId}/contacts'.format(tagId=tag_id)
        return self._post(endpoint, data=dumps(data))

    def remove_tag(self, tag_id, contact_id, **kwargs):
        """
        Remove a tag from a contact.
        :param tag_id: Integer tag ID.
        :param contact_id: Integer contact ID.
        :param kwargs:
        :return:
        """
        if not tag_id:
            raise KeapDataError("tag_id is required.")
        if not contact_id:
            raise KeapDataError("contact_id is required.")

        endpoint = 'tags/{tagId}/contacts/{contactId}'.format(tagId=tag_id, contactId=contact_id)
        return self._delete(endpoint, **kwargs)

    def achieve_goal(self, integration, callName, **kwargs):
        """
        Achieve an api goal
        :param integration: Integration name.
        :param callName: callname on the integration.
        :return:
        """
        if not integration:
            raise KeapDataError("integration is required.")
        if not callName:
            raise KeapDataError("callName is required.")

        if kwargs is not None:
            params = {}
            params.update(kwargs)
            endpoint = 'campaigns/goals/{integration_Name}/{call_Name}'.format(integration_Name=integration,
                                                                               call_Name=callName)
            return self._post(endpoint, json_data=params)
        raise KeapDataError("Data is required to achieve an api goal")

    def query_contact_custom_field(self, field_name, field_value, return_fields=None):
        """
        return id of contact matching query
        :param return_fields: fields to return from keap
        :param field_name: Name of the custom field in Keap. ex: _netsuiteid
        :param field_value: value to query on
        :return:
        """
        if not field_name:
            raise KeapDataError("field_name is required.")
        if not field_value:
            raise KeapDataError("field_value is required.")
        xmlrpc_client = self.get_xmlrpc_client()
        query = {
            f"{field_name}": f"{field_value}"
        }
        fields = ["Id"]
        if return_fields is not None:
            for field in return_fields:
                fields.append(field)
        return xmlrpc_client.DataService("query", "Contact", 1, 0, query, fields)



from setuptools import setup

setup(
    name='keap_flask',
    version='0.2.0',
    description='A Rest Client For Flask applications',
    author='will sexton',
    author_email='will@theapiguys.com',
    url='https://github.com/codinlikewilly/keap_flask',
    license='BSD 2-clause',
    packages=['keap_flask'],
    install_requires=['Flask',
                      'AuthLib',
                      'python-dateutil'
                      ],

    classifiers=[
        'License :: OSI Approved :: BSD License'
    ],
)
